# Coursework sheet 1: Simulation

# Solutions to questions 6, 7 and 8

# --------------------------------------------------
#                   Question 6
# --------------------------------------------------

# This function doesn't actually return the simulated values
# but only the acceptance rate, i.e. the number of proposals
# needed to generate nosam values in n dimensions divided by
# nosam.

mdrs <- function(n,nosam) {  # Multi-dimensional rejection sampling

  cnt <- 0
  for (i in 1:nosam) {
     repeat {
       u <- runif(n,-1,1)   # proposal
       cnt <- cnt+1
       if (runif(1)<exp(-sum(u*u)/2)) break
     }
  }
  cat("Dimension    :",n,"\n")
  cat("Sample size  :",nosam,"\n")
  cat("No proposals :",cnt,"\n")
  cat("Acc. rate    :",nosam/cnt,"\n")
  nosam/cnt
}

# Generate 1000 realisations in dimensions at most 20,
# 100 realisations in dimensions 21 to 40 and
# 10 realisations in dimensions 41 to 50. The numbers are
# chosen such that it took a couple of minutes to run
# the simulations on my computer. For higher precision of
# the estimates of the acceptance rates more simulations are needed.

nvec <- seq(1:50)
avec <- apply(matrix(1:20),1,mdrs,1000)
avec <- c(avec,apply(matrix(21:40),1,mdrs,100))
avec <- c(avec,apply(matrix(41:50),1,mdrs,10))

plot(nvec,avec,xlab='dimension',ylab='acc.rate')

# Easier to see the dependence on n on a log-scale.

plot(nvec,log(avec),xlab='dimension',ylab='log(acc.rate)')

# The acceptance rate seems to be an exponentially decaying
# function of the dimension, i.e. we get a straight line on
# the log-plot. Let us fit the regression equation
# acc(n) = exp(-n^a * b ) using the function lm.

lm(log(-log(avec)) ~ log(nvec))

# This gave in my case:
# a = 0.988
# b = exp(-1.825) = 0.1612,

# Plot the fit on top.

fvec <- exp(-nvec^0.988*exp(-1.825))
plot(nvec,avec,xlab='dimension',ylab='acc.rate')
lines(nvec,fvec)

# Again, better to do this on a log-scale.

fvec <- -nvec^0.988*exp(-1.825)
plot(nvec,log(avec),xlab='dimension',ylab='acc.rate')
lines(nvec,fvec)


# If we believe this regression we can predict that for n=100
# we need on average: 1/exp(-100^0.988*exp(-1.825)) = 4.2 million
# iterations to generate a single realisation!
# This clearly is not a very efficient algorithm.

# --------------------------------------------------
#                   Question 7
# --------------------------------------------------

#  C_h = {(u,v): 0 < u < 1, abs(v) < sqrt(3*u*(1-u))}
#  This is contained in the rectangle [0,1] x [-0.87,0.87]

rut3 <- function(nosam) {

  r <- NULL
  i <- 0
  while (i < nosam) {
    u <- runif(1)
    v <- runif(1,-0.87,0.87)
    if  (abs(v)<sqrt(3*u*(1-u))) {
      i <- i+1
      r[i] <- v/u
    }
  }
  r
}

r <- rut3(1000)
hist(r,prob=T)
box()

# The normalisation constant in the t3_density is:
#   sqrt(3)B(3/2,1/2) = sqrt(3) Gamma(3/2)Gamma(1/2)
# = sqrt(3)/2 Gamma(1/2)Gamma(1/2) = pi*sqrt(3)/2 = 2.720699

x <- seq(min(r),max(r),length=50)
y <- 1/(sqrt(3)*pi/2*(1+x^2/3)^2)
lines(x,y)

# Zoom in on [-5,5]

rtrunc <- r[abs(r)<5]
hist(rtrunc,prob=T)
box()
x <- seq(-5,5,length=50)
y <- 1/(sqrt(3)*pi/2*(1+x^2/3)^2)
lines(x,y)

# --------------------------------------------------
#                   Question 8
# --------------------------------------------------

# True value: P(Z>3) = 0.001349898

# Naive approach using Box-Muller to generate normal variates

n  <- 10000
x  <- sqrt(-2*log(runif(n)))*cos(2*pi*runif(n))
ix <- x>3
p  <- sum(ix)/n
sd <- sqrt(var(ix)/n)
ci <- c(p-1.96*sd,p+1.96*sd)
cat("True value : ",format(1-pnorm(3),digits=8),"\nEstimate   : ",p,"\nStd.dev    : ",format(sd,digits=6,scientific=F),"\n95% CI     : ",format(ci,digits=6,scientific=F),"\n")

# Simple improvement using the antithetic variables -x.
# Note the way the standard deviation is calculated

ix <- (x>3)/2+(-x>3)/2
p  <- sum(ix)/n
sd <- sqrt(var(ix)/n)
ci <- c(p-1.96*sd,p+1.96*sd)
cat("True value : ",format(1-pnorm(3),digits=8),"\nEstimate   : ",p,"\nStd.dev    : ",format(sd,digits=6,scientific=F),"\n95% CI     : ",format(ci,digits=6,scientific=F),"\n")


# Importance sampling using an Exponential(3) distribution shifted to [3,oo)
# This leads to an improvement in the standard deviation of almost a factor 100.
# The parameter of the Exponential is chosen such that we have the best local
# agreement at x=3 between the normal density and the exponential density.
# Try different values of a and see how it changes the standard deviation.

a  <- 3        # The parameter of the exponential
x  <- 3+log(runif(n))/(-a)
ix <- exp(-x^2/2+a*(x-3))/(a*sqrt(2*pi))
p  <- sum(ix)/n
sd <- sqrt(var(ix)/n)
ci <- c(p-1.96*sd,p+1.96*sd)
cat("True value : ",format(1-pnorm(3),digits=8),"\nEstimate   : ",format(p,digits=6,scientific=F),"\nStd.dev    : ",format(sd,digits=6,scientific=F),"\n95% CI     : ",format(ci,digits=6,scientific=F),"\n")

# We can then combine this with a control variable.
# Let W = c*exp(-b*(x-3)),
# where b and c are parameters we can choose and x has an Exp(a) dist. shifted to [3,oo).
# The point is that with this choice of W we can calculate its expectation easily:
#  E[W] = ac/(a+b).

a  <- 3       # The parameter of the exponential
b  <- 1
c  <- exp(-9/2)/(a*sqrt(2*pi))  # Then W and ix are the same for x=3
x  <- 3+log(runif(n))/(-a)
w  <- c*exp(-b*(x-3))
ix <- exp(-x^2/2+a*(x-3))/(a*sqrt(2*pi))
p  <- sum(ix-w)/n + a*c/(a+b)
sd <- sqrt(var(ix-w)/n)
ci <- c(p-1.96*sd,p+1.96*sd)
cat("True value : ",format(1-pnorm(3),digits=8),"\nEstimate   : ",format(p,digits=6,scientific=F),"\nStd.dev    : ",format(sd,digits=6,scientific=F),"\n95% CI     : ",format(ci,digits=6,scientific=F),"\n")


# Finally, never forget to see if there is an analytic approximation
# which is good enough for the problem at hand. In this case we have from
# Lemma 2 of Feller: "An introduction to probability theory and
# its applications" vol. I that for all x>0 the double inequality holds
#
#   (1/x-1/x^3)f(x) < 1 - Phi(x) < 1/x f(x),
#
# where f and Phi are the pdf and cdf respectively of a standard normal.

cat("True value : ",format(1-pnorm(3),digits=8),"\nLower bound: ",format((1/3-1/27)*exp(-9/2)/sqrt(2*pi),digits=6,scientific=F),"\nUpper bound: ",format(1/3*exp(-9/2)/sqrt(2*pi),digits=6,scientific=F),"\n")
