# Coursework sheet 2: Markov chain Monte Carlo

# Solutions to questions 1, 2, 3, 4 and 5

# --------------------------------------------------
#                   Question 1
# --------------------------------------------------

simmc <- function(theta,n) {

   x _ 0
   trace _ NULL
   sd _ sqrt(1-theta^2)
   for (i in 1:n) {
     x _ rnorm(1,theta*x,sd)
     trace _ cbind(trace,x)
   }
   trace
}

t1 _ simmc(0.5,10000)
plot(t1[1:100],type='l')   # Trace plot of first 100 iterations

hist(t1,probability=T)     # Plot histogram 
x_seq(-4,4,len=100)        # with N(0,1) density superimposed
lines(x,dnorm(x))

# --------------------------------------------------
#                   Questions 2 and 3
# --------------------------------------------------

mhn <- function(n,init,a,tau) {

   x <- init
   xvec <- x
   for (i in 2:n) {
     y <- rnorm(1,a*x,tau)
     acc <- min(exp(-(y^2-x^2)/2-( (x-a*y)^2 - (y-a*x)^2 )/(2*tau^2)),1)
     if (runif(1)<acc) { x <- y }
     xvec <- c(xvec,x)
   }
   xvec
}

x1 <- mhn(1000,0,0,0.1)
x2 <- mhn(1000,0,0,0.5)
x3 <- mhn(1000,0,0,1)
x4 <- mhn(1000,0,0,2)

par(mfrow=c(2,2))
plot(x1,type='l')
title('a = 0, tau = 0.1')
plot(x2,type='l')
title('a = 0, tau = 0.5')
plot(x3,type='l')
title('a = 0, tau = 1')
plot(x4,type='l')
title('a = 0, tau = 2')

x1 <- mhn(1000,0,0.1,0.1)
x2 <- mhn(1000,0,0.1,0.5)
x3 <- mhn(1000,0,0.1,1)
x4 <- mhn(1000,0,0.1,2)

par(mfrow=c(2,2))
plot(x1,type='l')
title('a = 0.1, tau = 0.1')
plot(x2,type='l')
title('a = 0.1, tau = 0.5')
plot(x3,type='l')
title('a = 0.1, tau = 1')
plot(x4,type='l')
title('a = 0.1, tau = 2')

x1 <- mhn(1000,0,0.5,0.1)
x2 <- mhn(1000,0,0.5,0.5)
x3 <- mhn(1000,0,0.5,1)
x4 <- mhn(1000,0,0.5,2)

par(mfrow=c(2,2))
plot(x1,type='l')
title('a = 0.5, tau = 0.1')
plot(x2,type='l')
title('a = 0.5, tau = 0.5')
plot(x3,type='l')
title('a = 0.5, tau = 1')
plot(x4,type='l')
title('a = 0.5, tau = 2')

x1 <- mhn(1000,0,0.9,0.1)
x2 <- mhn(1000,0,0.9,0.5)
x3 <- mhn(1000,0,0.9,1)
x4 <- mhn(1000,0,0.9,2)

par(mfrow=c(2,2))
plot(x1,type='l')
title('a = 0.9, tau = 0.1')
plot(x2,type='l')
title('a = 0.9, tau = 0.5')
plot(x3,type='l')
title('a = 0.9, tau = 1')
plot(x4,type='l')
title('a = 0.9, tau = 2')

# Not surprisingly, a=0 and tau=1 is the best
# because in this case we get an i.i.d. sample from the target!
# Small deviations from this is of course also good,
# while tau=0.1 or 0.5 leads to poor mixing in all combinations with a.


# --------------------------------------------------
#                   Question 4
# --------------------------------------------------

gsn <- function(n,init,rho) {

  x <- NULL
  y <- NULL
  x[1] <- init[1]
  y[1] <- init[2]
  for (i in 2:n) {
    y[i] <- rnorm(1,rho*x[i-1],sqrt(1-rho^2))
    x[i] <- rnorm(1,rho*y[i],sqrt(1-rho^2))
  }
  cbind(x,y)
}

xy1 <- gsn(1000,c(3,3),0.7)
xy2 <- gsn(1000,c(3,3),0.9)
xy3 <- gsn(1000,c(3,3),0.95)
xy4 <- gsn(1000,c(3,3),0.99)

par(mfrow=c(2,2))
plot(xy1)
title('rho = 0.7')
plot(xy2)
title('rho = 0.9')
plot(xy3)
title('rho = 0.95')
plot(xy4)
title('rho = 0.99')

par(mfrow=c(2,2))
plot(xy1[,1],type='l')
title('rho = 0.7')
plot(xy2[,1],type='l')
title('rho = 0.9')
plot(xy3[,1],type='l')
title('rho = 0.95')
plot(xy4[,1],type='l')
title('rho = 0.99')

par(mfrow=c(2,2))
temp_acf(xy1[,1])
temp_acf(xy2[,1])
temp_acf(xy3[,1],lag.max=1000)
temp_acf(xy4[,1],lag.max=1000)


# --------------------------------------------------
#                   Question 5
# --------------------------------------------------

x1 <- xy1[,1]
x2 <- xy2[,1]
x3 <- xy3[,1]
x4 <- xy4[,1]

# Batch mean estimation of variance based on batches of size m

bmest <- function(x,m) {

  n <- length(x)
  l <- n/m
  cat("Sample size:",n,"\nBatch size :",m,"\nNo batches :",l,"\n");
  b <- NULL
  for (i in 1:l) {
    b[i] <- mean(x[((i-1)*m+1):(i*m)])
  }
  m*var(b)
}

x <- x1
sd <- sqrt(bmest(x,5)/length(x))
ci <- c(mean(x)-1.96*sd,mean(x)+1.96*sd)
wrongsd <- sqrt(var(x)/n)
wrongci <- c(mean(x)-1.96*wrongsd,mean(x)+1.96*wrongsd)
cat("Estimate :",format(mean(x),digits=6,scientific=F),"\n95% CI   :",format(ci,digits=6,scientific=F),"\nNaive CI :",format(wrongci,digits=6,scientific=F),"\n")

x <- x2
sd <- sqrt(bmest(x,15)/length(x))
ci <- c(mean(x)-1.96*sd,mean(x)+1.96*sd)
wrongsd <- sqrt(var(x)/n)
wrongci <- c(mean(x)-1.96*wrongsd,mean(x)+1.96*wrongsd)
cat("Estimate :",format(mean(x),digits=6,scientific=F),"\n95% CI   :",format(ci,digits=6,scientific=F),"\nNaive CI :",format(wrongci,digits=6,scientific=F),"\n")

x <- x3
sd <- sqrt(bmest(x,100)/length(x))
ci <- c(mean(x)-1.96*sd,mean(x)+1.96*sd)
wrongsd <- sqrt(var(x)/n)
wrongci <- c(mean(x)-1.96*wrongsd,mean(x)+1.96*wrongsd)
cat("Estimate :",format(mean(x),digits=6,scientific=F),"\n95% CI   :",format(ci,digits=6,scientific=F),"\nNaive CI :",format(wrongci,digits=6,scientific=F),"\n")

x <- x4
sd <- sqrt(bmest(x,200)/length(x))
ci <- c(mean(x)-1.96*sd,mean(x)+1.96*sd)
wrongsd <- sqrt(var(x)/n)
wrongci <- c(mean(x)-1.96*wrongsd,mean(x)+1.96*wrongsd)
cat("Estimate :",format(mean(x),digits=6,scientific=F),"\n95% CI   :",format(ci,digits=6,scientific=F),"\nNaive CI :",format(wrongci,digits=6,scientific=F),"\n")

# Hopefully, all the right CIs include zero and
# at least one of the wrong CIs does not include zero.
# You can also play with the value of m and see how the CI changes.
