# Coursework sheet 3: Bootstrap and jackknife

# Solutions to questions 4 and 5
# --------------------------------------------------

# The jackknife function from p.72 in the notes
# here renamed jackknife1 to avoid conflict with
# the Splus function jackknife.

jackknife1 <- function(x,theta,...) {

  z  <- NULL
  n  <- length(x)
  jd <- NULL
  for (i in 1:n) {
    jd[i] <- theta(x[-i],...)
  }
  est     <- theta(x,...)
  z$est   <- est
  z$distn <- jd
  z$bias  <- (mean(jd)-est)*(n-1)
  z$se    <- sqrt(var(jd)*(n-1)^2/n)
  z
}

# --------------------------------------------------
#                   Question 4
# --------------------------------------------------

# Estimation of the median in the salamander data
# --------------------------------------------------

hist(sal.dat)
z <- bootstrap1(sal.dat,200,median)
hist(z$distn)
z$est
z$bias
z$se
u <- jackknife1(sal.dat,median)
hist(u$distn)
u$est
u$bias
u$se

# Note that the jackknife estimate of variance is 0!
# This of course is useless information and tells us
# nothing about the true variation of the estimate of the median.
# The reason is that by deleting only one observation at
# a time we never change the median. The inflation factors
# of (n-1) and (n-1)^2/n, respectively, which are introduced
# to correct for this phenomenon of under-variation in
# the jackknife sample do not help in this case.

# Estimation of the correlation in law.dat
# --------------------------------------------------

plot(law.dat)
z <- bootstrap1(1:15,200,theta1,law.dat)
hist(z$distn)
z$est
z$bias
z$se
u <- jackknife1(1:15,theta1,law.dat)
u$est
u$bias
u$se

# In this case we get a better agreement between the
# two methods.

# Comparing two populations
# --------------------------------------------------

par(mfrow=c(1,2))
hist(ph1.dat)
title('Historical')
hist(ph2.dat)
title('Current')
z <- bootstrap2(ph1.dat,ph2.dat,200,theta2,length(ph1.dat))
par(mfrow=c(1,1))
hist(z$distn,xlab='Difference in sample medians')
z$est
z$bias
z$se

# Evidence of some but not much bias.

# Like in the first example it is not clear how
# to get useful jackknife estimates in this case.


# Time serie
# --------------------------------------------------

theta3noplot <- function(x, resid, beta, z1)
 {
        y <- NULL
        y[1] <- z1
        for(i in 1:length(resid)) {
                y[i + 1] <- beta * y[i] + resid[x[i]]
        }
        ar.fit <- ar(y, aic = F, ord = 1)
        ar.fit$ar
}


plot(hormone.dat,type='l')
afit <- ar(hormone.dat,aic=F,ord=1)
resid <- as.vector(afit$resid[2:48])
beta  <- as.numeric(afit$ar)
z1    <- hormone.dat[1]
hist(resid)
z <- bootstrap1(1:47,200,theta3noplot,resid,beta,z1)
hist(z$distn)
z$est
z$bias
z$se

# Eight bootstrapped series and the orginial series

par(mfrow=c(3,3))
z <- bootstrap1(1:47,8,theta3,resid,beta,z1)

# Jackknife estimates not immediately applicable.

# Non-parametric regression
# --------------------------------------------------

bootstrap3 <- function(x, nboot, data)
{
     plot(mc.dat)
     for (i in 1:nboot) {
        idx   <- sample(x, size = length(x), replace = T)
        mc.lo <- loess(data[idx, 2] ~ data[idx, 1], span = 1/3)
        y <- predict.loess(mc.lo, 1:60)
       lines(1:60, y)
    }
    mc.lo <- loess(data[,2] ~ data[,1],span=1/3)
    y <- predict.loess(mc.lo,1:60)
    lines(1:60,y,col=8)
}

par(mfrow=c(1,1))
bootstrap3(1:133,20,mc.dat)

# Jackknife estimates obtained by deleting one observation
# in turn will greatly under-estimate the variability and
# not provide useful information.

# Regression
# --------------------------------------------------

# The dataset used is this example is called wood.dat

# Method 1: bootstrap the pairs

plot(wood.dat)
ls <- lsfit(wood.dat[,1],wood.dat[,2])
abline(ls)
ls$coef

z <- bootstrap1(1:36,10,theta5,wood.dat)
abline(ls,col=8)

theta5noplot <- function(x, xdata)
{
        ls <- lsfit(xdata[x, 1], xdata[x, 2])
        as.vector(ls$coef)
}


z <- bootstrap1(1:36,200,theta5noplot,wood.dat)
b0 <- z$distn[1,]
b1 <- z$distn[2,]
par(mfrow=c(1,2))
hist(b0)
hist(b1)
mean(b0)
sqrt(var(b0))
mean(b1)
sqrt(var(b1))

# Jackknife estimates not immediately applicable.


# Method 2: bootstrap the residuals

par(mfrow=c(1,1))
plot(wood.dat)
ls <- lsfit(wood.dat[,1],wood.dat[,2])
abline(ls)
ls$coef
resid <- as.vector(ls$resid)
woodfit <- cbind(wood.dat[,1],wood.dat[,2]-resid)

z <- bootstrap1(resid,10,theta6,woodfit)
abline(ls,col=8)

# Note: bootstrap1 should be called with woodfit NOT wood.dat
#       as in the notes p. 69. Woodfit represents the linear fit
#       and to this we add the bootstrapped residuals to get y.

theta6noplot <- function(x, xdata)
{
        y <- x + xdata[, 2]
        ls <- lsfit(xdata[, 1], y)
        as.vector(ls$coef)
}

z <- bootstrap1(resid,200,theta6noplot,woodfit)
b0 <- z$distn[1,]
b1 <- z$distn[2,]
par(mfrow=c(1,2))
hist(b0)
hist(b1)
mean(b0)
sqrt(var(b0))
mean(b1)
sqrt(var(b1))

# Jackknife estimates not immediately applicable.

# --------------------------------------------------
#                   Question 5
# --------------------------------------------------

par(mfrow=c(1,1))
plot(fuel)
flo <- loess(fuel[,2] ~ fuel[,1],span=1/3)
y <- predict.loess(flo,seq(0.535,1.232,length=40))
lines(seq(0.535,1.232,length=40),y)

theta7 <- function(x,xdata) {
  flo <- loess(xdata[x,2] ~ xdata[x,1],span=1/3)
  y <- predict.loess(flo,seq(0.535,1.232,length=40))
  lines(seq(0.535,1.232,length=40),y)
}

bootstrap1(1:88,20,theta4,fuel)
lines(seq(0.535,1.232,length=40),y,col=8)

# A pointwise 90% confidence intervals can be produced by
# taking the upper and lower 5% quantile of the lines
# for each x value.

bootstrap4 <- function(x, nboot, data)
{
    plot(data)
    ymat <- NULL
    for (i in 1:nboot) {
        idx   <- sample(x, size = length(x), replace = T)
        flo <- loess(data[idx, 2] ~ data[idx, 1], span = 1/3)
        y <- predict.loess(flo,seq(0.535,1.232,length=40))
        ymat <- rbind(ymat,y)
    }
    flo <- loess(data[,2] ~ data[,1],span=1/3)
    y <- predict.loess(flo,seq(0.535,1.232,length=40))
    lines(seq(0.535,1.232,length=40),y,col=8)

    ci <- apply(zmat,2,quantile,probs=c(0.05,0.9),na.rm=T)
    lines(seq(0.535,1.232,length=40),ci[1,])
    lines(seq(0.535,1.232,length=40),ci[2,])
}

z <- bootstrap4(1:88,100,fuel)
